@@ -1,6 +1,8 @@
 #include "lambda.H"
 #include "var.H"
-#include "operator.H"
+
+#include "range/v3/all.hpp"
+namespace views = ranges::views;
 
 using std::string;
 using std::vector;
@@ -33,11 +35,24 @@ expression_ref lambda_quantify(int var_index, const expression_ref& R)
     return lambda_quantify(var(var_index), R);
 }
 
-expression_ref lambda_expression(const Operator& O)
+expression_ref lambda_quantify(const vector<expression_ref>& vars, expression_ref& body)
+{
+    auto L = body;
+    for(auto& v: vars | views::reverse)
+        L = lambda_quantify(v, L);
+    return L;
+}
+
+expression_ref lambda_quantify(const std::vector<var>& vars, const expression_ref& body)
+{
+    auto L = body;
+    for(auto& v: vars | views::reverse)
+        L = lambda_quantify(v, L);
+    return L;
+}
+
+expression_ref lambda_n(const Object& O, int n)
 {
-    int n = O.n_args();
-    assert(n != -1);
-  
     expression_ref R;
     if (n == 0)
 	R = expression_ref(O.clone());
@@ -55,7 +70,7 @@ expression_ref lambda_expression(const Operator& O)
     return R;
 }
 
-expression_ref lambda_expression(const Operator& O);
+
 expression_ref lambda_quantify(int var_index, const expression_ref& E);
 expression_ref lambda_quantify(const expression_ref& var, const expression_ref& E);
 
