@@ -21,13 +21,11 @@
 #include "util/range.H"
 #include "util/rng.H"
 #include "util/log-level.H"
+#include "util/settings.H"       // for get_setting_or( )
 #include <algorithm>
-#include "mcmc.H"
-#include "proposals.H"
 #include "dp/3way.H"
 #include "util/permute.H"
 #include "alignment/alignment-util.H"
-#include "mcmc/logger.H"
 
 using MCMC::MoveStats;
 
@@ -35,26 +33,29 @@ using std::valarray;
 using std::vector;
 using std::string;
 using std::map;
+using std::optional;
 
-void slide_node_move(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+
+void slide_node_move(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     slide_node(P,Stats,b);
 }
 
-void change_branch_length_move(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+void change_branch_length_move(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     change_branch_length(P,Stats,b);
 }
 
-void change_branch_length_multi_move(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+void change_branch_length_multi_move(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     change_branch_length_multi(P,Stats,b);
 }
 
-void sample_tri_one(owned_ptr<Model>& P, MoveStats&,int b) 
+void sample_tri_one(owned_ptr<context>& P, MoveStats&,int b) 
 {
     Parameters* PP = P.as<Parameters>();
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
 
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -62,21 +63,22 @@ void sample_tri_one(owned_ptr<Model>& P, MoveStats&,int b)
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     tri_sample_alignment(*PP,node1,node2);
 }
 
-void sample_tri_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+void sample_tri_branch_one(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     Parameters* PP = P.as<Parameters>();
 
     MCMC::Result result(2);
 
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
   
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -84,7 +86,7 @@ void sample_tri_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     const double sigma = 0.3/2;
@@ -100,10 +102,11 @@ void sample_tri_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
     Stats.inc("sample_tri_branch",result);
 }
 
-void sample_cube_one(owned_ptr<Model>& P, MoveStats&,int b) 
+void sample_cube_one(owned_ptr<context>& P, MoveStats&,int b) 
 {
     Parameters* PP = P.as<Parameters>();
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
 
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -111,21 +114,22 @@ void sample_cube_one(owned_ptr<Model>& P, MoveStats&,int b)
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     cube_sample_alignment(*PP,node1,node2);
 }
 
-void sample_cube_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+void sample_cube_branch_one(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     Parameters* PP = P.as<Parameters>();
 
     MCMC::Result result(2);
 
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
   
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -133,7 +137,7 @@ void sample_cube_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     const double sigma = 0.3/2;
@@ -150,15 +154,16 @@ void sample_cube_branch_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
 }
 
 
-void sample_parameter_and_alignment_on_branch(owned_ptr<Model>& P, MoveStats& Stats, int b, const Proposal& proposal)
+void sample_parameter_and_alignment_on_branch(owned_ptr<context>& P, MoveStats& Stats, int b, const Proposal& proposal)
 {
     Parameters* PP = P.as<Parameters>();
 
     MCMC::Result result(1);
 
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
 
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -166,7 +171,7 @@ void sample_parameter_and_alignment_on_branch(owned_ptr<Model>& P, MoveStats& St
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     if (tri_sample_alignment_and_parameter(*PP, node1, node2, proposal))
@@ -180,15 +185,16 @@ void sample_parameter_and_alignment_on_branch(owned_ptr<Model>& P, MoveStats& St
 }
 
 
-void sample_tri_branch_type_one(owned_ptr<Model>& P, MoveStats& Stats,int b) 
+void sample_tri_branch_type_one(owned_ptr<context>& P, MoveStats& Stats,int b) 
 {
     Parameters* PP = P.as<Parameters>();
 
     MCMC::Result result(1);
 
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
     auto t = PP->t();
+    assert(is_degree3_edge(t,b));
 
     int node1 = t.target(t.undirected(b));
     int node2 = t.source(t.undirected(b));
@@ -196,7 +202,7 @@ void sample_tri_branch_type_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
     if (uniform() < 0.5)
         std::swap(node1,node2);
 
-    if (node1 < t.n_leaves())
+    if (t.degree(node1) != 3)
         std::swap(node1,node2);
     
     if (tri_sample_alignment_branch_model(*PP,node1,node2)) {
@@ -207,42 +213,48 @@ void sample_tri_branch_type_one(owned_ptr<Model>& P, MoveStats& Stats,int b)
 }
 
 
-void sample_alignments_one(owned_ptr<Model>& P, MoveStats& Stats, int b)
+void sample_alignments_one(owned_ptr<context>& P, MoveStats& Stats, int b)
 {
+    double alignment_plus_branch_length_fraction = get_setting_or("alignment_plus_branch_length_fraction",0.01);
     Parameters* PP = P.as<Parameters>();
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
-    if (uniform() < 0.01)
+    if (PP->t().has_branch_lengths() and uniform() < alignment_plus_branch_length_fraction)
         alignment_slice_sample_branch_length(P, Stats, b);
     else
         sample_alignment(*PP,b);
 }
 
-void sample_node_move(owned_ptr<Model>& P, MoveStats&,int node) 
+void sample_node_move(owned_ptr<context>& P, MoveStats&,int node) 
 {
     Parameters* PP = P.as<Parameters>();
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
     sample_node(*PP,node);
 }
 
-void sample_two_nodes_move(owned_ptr<Model>& P, MoveStats&,int n0) 
+void sample_A5_move(owned_ptr<context>& P, MoveStats&,int n0) 
 {
     Parameters* PP = P.as<Parameters>();
-    assert(PP->variable_alignment()); 
+//    assert(PP->variable_alignment()); 
 
-    vector<int> nodes = A3::get_nodes_random(PP->t(),n0);
-    int n1 = -1;
+    auto t = PP->t();
+    assert(t.degree(n0) == 3);
+
+    vector<int> nodes = A3::get_nodes_random(t,n0);
+    optional<int> n1;
     for(int i=1;i<nodes.size();i++)
-        if ((PP->t()).is_internal_node( nodes[i] )) {
+        if (t.degree(nodes[i]) == 3)
+	{
             n1 = nodes[i];
             break;
-        }
-    assert(n1 != 1);
+        };
 
-    int b = PP->t().undirected(PP->t().find_branch(n0,n1));
+    if (not n1) return;
+    
+    int b = PP->t().undirected(PP->t().find_branch(n0, *n1));
 
-    sample_two_nodes(*PP,b);
+    sample_A5(*PP,b);
 }
 
 // cost[b] should be the cost to visit all the branches after b.
@@ -251,58 +263,90 @@ void sample_two_nodes_move(owned_ptr<Model>& P, MoveStats&,int n0)
 //   be on one of the endpoints of the branch
 // * the cost of a non-leaf branch with branches_out={l,r} is 2*cost[l]+cost[r]
 //   if we visit the subtree in front of l first.
-vector<int> get_cost(const TreeInterface& t) {
-    vector<int> cost(t.n_branches()*2,-1);
+std::unordered_map<int,int> get_cost(const TreeInterface& t)
+{
+    std::unordered_map<int,int> cost;
+
     vector<int> stack1; stack1.reserve(t.n_branches()*2);
     vector<int> stack2; stack2.reserve(t.n_branches()*2);
-    for(int i=0;i<t.n_leaves();i++) {
-        int b = t.reverse(i);
-        cost[b] = 0;
-        stack1.push_back(b);
+
+    for(auto b: t.leaf_branches())
+    {
+        // Leaf branches point away from the leaf.
+        // So b2 points toward the leaf.
+        int b2 = t.reverse(b);
+        cost.insert({b2,0});
+        stack1.push_back(b2);
     }
+
+    if (t.n_nodes() == 2)
+        return cost;
     
-    while(not stack1.empty()) {
+    while(not stack1.empty())
+    {
         // fill 'stack2' with branches before 'stack1'
         stack2.clear();
-        for(int i=0;i<stack1.size();i++)
-            t.append_branches_before(stack1[i], stack2);
+        for(int b: stack1)
+            t.append_branches_before(b, stack2);
 
         // clear 'stack1'
+        // these branches are already done -- we just needed to process the branches before them
         stack1.clear();
 
-        for(int i=0;i<stack2.size();i++) {
-            vector<int> children = t.branches_after(stack2[i]);
+        for(int b: stack2)
+        {
+            // Skip b if we already have already processed it.
+            if (cost.contains(b)) continue;
 
-            assert(children.size() == 2);
-            int cost_l = cost[children[0]];
-            int cost_r = cost[children[1]];
-            if (cost_l != -1 and cost_r != -1) {
-                if (not t.is_leaf_branch(children[0])) cost_l++;
+            auto children = t.branches_after(b);
 
-                if (not t.is_leaf_branch(children[1])) cost_r++;
+            bool all_children_done = true;
+            int total_cost = 0;
+            int worst_cost = 0;
+            for(int b: children)
+            {
+                if (not cost.contains(b))
+                {
+                    all_children_done = false;
+                    break;
+                }
 
-                if (cost_l > cost_r)
-                    std::swap(cost_l,cost_r);
+                int child_cost = cost.at(b);
 
-                cost[stack2[i]] = 2*cost_l + cost_r;
-                stack1.push_back(stack2[i]);
+                // It costs 1 more to cross b if it is not a leaf branch.
+                if (not t.is_leaf_branch(b)) child_cost++;
+
+                // Find the highest cost branch, which we will traverse only once.
+                worst_cost = std::max(worst_cost, child_cost);
+
+                total_cost += child_cost;
+            }
+
+            if (all_children_done)
+            {
+                // We traverse all branches twice, except for the worst one
+                cost.insert({b, 2*total_cost - worst_cost});
+
+                // Notify that b is now done, and we can try (or re-try) handling
+                // the branches that depend on it.
+                stack1.push_back(b);
             }
         }
     }
-  
+
     // check that all the costs have been calculated
-    for(int i=0;i<cost.size();i++)
-        assert(cost[i] != -1);
+    for(int b: t.directed_branches())
+        assert(cost.contains(b));
 
     return cost;
 }
 
-vector<int> get_distance(const TreeInterface& t, int n)
+std::unordered_map<int,int> get_distance(const TreeInterface& t, int n)
 {
-    vector<int> D(t.n_nodes(),-1);
+    std::unordered_map<int,int> D;
     D[n] = 0;
 
-    vector<int> branches = t.branches_out(n);
+    auto branches = t.branches_out(n);
     int d = 0;
     int i = 0;
     while(i<branches.size())
@@ -374,12 +418,13 @@ vector<int> walk_tree_path_toward_and_away(const TreeInterface& t, int node)
 
 vector<int> walk_tree_path(const TreeInterface& t, int root)
 {
-    vector<int> cost = get_cost(t);
+    auto cost = get_cost(t);
 
-    vector<int> D = get_distance(t,root);
-    vector<int> tcost = cost;
-    for(int i=0;i<cost.size();i++)
-        tcost[i] += D[t.target(i)];
+    auto D = get_distance(t,root);
+
+    auto tcost = cost;
+    for(auto& [b,c]: tcost)
+        c += D.at(t.target(b));
 
     vector<int> b_stack;
     b_stack.reserve(t.n_branches());
@@ -388,17 +433,23 @@ vector<int> walk_tree_path(const TreeInterface& t, int root)
     vector<int> children;
     children.reserve(3);
 
-    // get a leaf with minimum 'tcost'
-    int leaf = 0;
-    leaf = myrandom(t.n_leaves());
-    for(int b=0;b<t.n_leaves();b++)
-        if (tcost[b] < tcost[leaf])
-            leaf = b;
+    // get a leaf branch with minimum 'tcost'
+    std::optional<int> best_leaf_branch;
+    for(auto b: t.leaf_branches())
+    {
+        if (not best_leaf_branch)
+            best_leaf_branch = b;
+        else if (tcost.at(b) < tcost.at(best_leaf_branch.value()))
+            best_leaf_branch = b;
+    }
 
-    assert(t.source(leaf) == leaf);
-    b_stack.push_back(leaf);
+    // If the tree has no branches, then quit here.
+    if (not best_leaf_branch) return {};
 
-    while(not b_stack.empty()) {
+    b_stack.push_back(*best_leaf_branch);
+
+    while(not b_stack.empty())
+    {
         // pop stack into list
         branches.push_back(b_stack.back());
         b_stack.pop_back();
@@ -413,13 +464,13 @@ vector<int> walk_tree_path(const TreeInterface& t, int root)
             ;
         else {
             if (children.size() == 2) {
-                if (cost[children[0]] < cost[children[1]])
+                if (cost.at(children[0]) < cost.at(children[1]))
                     std::swap(children[0],children[1]);
             }
             else
                 std::abort();
         }
-      
+
         // put children onto the stack
         for(int b: children)
             b_stack.push_back(b);
@@ -434,11 +485,12 @@ vector<int> walk_tree_path(const TreeInterface& t, int root)
     return branches2;
 }
 
-void sample_branch_length_(owned_ptr<Model>& P,  MoveStats& Stats, int b)
+void sample_branch_length_(owned_ptr<context>& P,  MoveStats& Stats, int b)
 {
+    if (log_verbose >= 3) std::cerr<<"\n\n[sample_branch_length_]\n";
     //std::clog<<"Processing branch "<<b<<" with root "<<P.subst_root()<<endl;
 
-    double slice_fraction = P->load_value("branch_slice_fraction",0.9);
+    double slice_fraction = get_setting_or("branch_slice_fraction",0.9);
 
     bool do_slice = (uniform() < slice_fraction);
     if (do_slice)
@@ -452,7 +504,7 @@ void sample_branch_length_(owned_ptr<Model>& P,  MoveStats& Stats, int b)
     if (uniform() < 0.5)
         e = e.reverse();
 
-    if (t.is_leaf_node(e.node2))
+    if (t.degree(e.node2) != 3)
         e = e.reverse();
 
     // NOTE! This pointer might be invalidated after the tree is changed by MH!
@@ -460,7 +512,7 @@ void sample_branch_length_(owned_ptr<Model>& P,  MoveStats& Stats, int b)
 
     // FIXME - this might move the accumulator off of the current branch (?)
     // TEST and Check Scaling of # of branches peeled
-    if (t.n_nodes() > 2)
+    if (is_degree3_edge(t,e))
     {
         if (uniform() < 0.5)
             slide_node(P, Stats, t.find_branch(e));
@@ -474,18 +526,16 @@ void sample_branch_length_(owned_ptr<Model>& P,  MoveStats& Stats, int b)
     }
 }
 
-void walk_tree_sample_NNI_and_branch_lengths(owned_ptr<Model>& P, MoveStats& Stats) 
+void walk_tree_sample_NNI_and_branch_lengths(owned_ptr<context>& P, MoveStats& Stats) 
 {
     Parameters& PP = *P.as<Parameters>();
-    vector<int> branches = walk_tree_path(PP.t(), PP[0].subst_root());
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
 
-    for(int i=0;i<branches.size();i++)
+    for(int b: branches)
     {
-        int b = branches[i];
-
         double U = uniform();
 
-        if (U < 0.1)
+        if (PP.t().has_branch_lengths() and U < 0.1)
             slice_sample_branch_length(P,Stats,b);
 
         if (PP.t().is_internal_branch(b)) 
@@ -500,20 +550,50 @@ void walk_tree_sample_NNI_and_branch_lengths(owned_ptr<Model>& P, MoveStats& Sta
                 two_way_NNI_sample(P,Stats,b);
         }
 
-        if (U > 0.9)
+        if (PP.t().has_branch_lengths() and U > 0.9)
             slice_sample_branch_length(P,Stats,b);
     }
 }
 
 
-void walk_tree_sample_NNI(owned_ptr<Model>& P, MoveStats& Stats)
+void walk_time_tree_sample_NNI_and_node_times(owned_ptr<context>& P, MoveStats& Stats)
+{
+    if (log_verbose >= 3) std::cerr<<"\n\n[walk_time_tree_sample_NNI_and_node_times]\n";
+    Parameters& PP = *P.as<Parameters>();
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
+
+    for(int b: branches)
+    {
+	int n1 = PP.t().source(b);
+	int n2 = PP.t().target(b);
+        double U = uniform();
+
+        if (U < 0.1)
+	{
+            slice_sample_node_time(P,Stats,n1);
+            slice_sample_node_time(P,Stats,n2);
+	}
+
+	three_way_time_tree_NNI_sample(P,Stats,b);
+
+        if (U > 0.9)
+	{
+            slice_sample_node_time(P,Stats,n2);
+            slice_sample_node_time(P,Stats,n1);
+	}
+    }
+}
+
+
+void walk_tree_sample_NNI(owned_ptr<context>& P, MoveStats& Stats)
 {
     Parameters& PP = *P.as<Parameters>();
-    vector<int> branches = walk_tree_path(PP.t(), PP[0].subst_root());
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
 
-    for(int i=0;i<branches.size();i++) 
+    for(int b: branches)
     {
-        int b = branches[i];
+        if (log_verbose >= 3)
+            std::cerr<<" <walk_tree_sample_NNI: branch "<<b<<">\n";
         if (uniform() < 0.95)
             three_way_topology_sample(P,Stats,b);
         else
@@ -522,16 +602,17 @@ void walk_tree_sample_NNI(owned_ptr<Model>& P, MoveStats& Stats)
 }
 
 
-void walk_tree_sample_NNI_and_A(owned_ptr<Model>& P, MoveStats& Stats) 
+void walk_tree_sample_NNI_and_A(owned_ptr<context>& P, MoveStats& Stats) 
 {
-    double NNI_A_fraction = P->load_value("NNI+A_fraction",0.01);
+    double NNI_A_fraction = get_setting_or("NNI+A_fraction",0.01);
 
     Parameters& PP = *P.as<Parameters>();
-    vector<int> branches = walk_tree_path(PP.t(), PP[0].subst_root());
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
 
-    for(int i=0;i<branches.size();i++) 
+    for(int b: branches)
     {
-        int b = branches[i];
+        if (log_verbose >= 3)
+            std::cerr<<" <walk_tree_sample_NNI_and_A: branch "<<b<<">\n";
         if (uniform() < NNI_A_fraction)
             three_way_topology_and_alignment_sample(P,Stats,b);
         else
@@ -543,20 +624,20 @@ void walk_tree_sample_NNI_and_A(owned_ptr<Model>& P, MoveStats& Stats)
 }
 
 
-void walk_tree_sample_alignments(owned_ptr<Model>& P, MoveStats& Stats) 
+void walk_tree_sample_alignments(owned_ptr<context>& P, MoveStats& Stats) 
 {
     Parameters& PP = *P.as<Parameters>();
-    vector<int> branches = walk_tree_path(PP.t(), PP[0].subst_root());
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
 
-    double cube_fraction = P->load_value("cube_fraction",0.00);
+    double cube_fraction = get_setting_or("cube_fraction",0.00);
 
-    for(int i=0;i<branches.size();i++) 
+    for(int b: branches)
     {
-        int b = branches[i];
-
+        if (log_verbose >= 3)
+            std::cerr<<" <walk_tree_sample_alignments: branch "<<b<<">\n";
         //    std::clog<<"Processing branch "<<b<<" with root "<<P.subst_root()<<endl;
 
-        if ((uniform() < 0.15) and (PP.t().n_leaves() >2))
+        if ((uniform() < 0.15) and is_degree3_edge(PP.t(),b))
         {
             // FIXME: don't call sample_parameter_and_alignment_on_branch( ): something is wrong.
             if (uniform() < 0.5 or true)
@@ -578,20 +659,21 @@ void walk_tree_sample_alignments(owned_ptr<Model>& P, MoveStats& Stats)
 
 
 // FIXME: Realign from tips basically fails because the distance between sequences is too small!
-void realign_from_tips(owned_ptr<Model>& P, MoveStats& Stats) 
+void realign_from_tips(owned_ptr<context>& P, MoveStats& Stats) 
 {
-    int AL0 = MCMC::alignment_length(*P.as<Parameters>());
+    int AL0 = alignment_length(*P.as<Parameters>());
 
     if (log_verbose>=3) std::cerr<<"realign_from_tips: |A0| = "<<AL0<<"\n";
     Parameters& PP = *P.as<Parameters>();
     auto t = PP.t();
-    int toward_node = uniform(t.n_nodes()>2?t.n_leaves():0, t.n_nodes()-1);
+    int toward_node = (t.n_leaves() > 2) ? uniform_element(t.internal_nodes()) : uniform_element(t.nodes());
+
     vector<int> branches = walk_tree_path_toward_and_away(t, toward_node);
 
     for(int b: branches)
     {
-        sample_branch_length_(P,Stats,b);
         auto t = P.as<Parameters>()->t();
+        if (t.has_branch_lengths() and t.can_set_branch_length(b)) sample_branch_length_(P,Stats,b);
         int node1 = t.source(b);
         int node2 = t.target(b);
         if (log_verbose >=4)
@@ -617,7 +699,7 @@ void realign_from_tips(owned_ptr<Model>& P, MoveStats& Stats)
             }
             std::cerr<<"\n";
         }
-        if (node2 >= t.n_leaves())
+        if (t.degree(node2) == 3)
         {
             if (log_verbose >=3) std::cerr<<"     Performing 3-way alignment\n";
             tri_sample_alignment(*P.as<Parameters>(), node2, node1);
@@ -627,7 +709,7 @@ void realign_from_tips(owned_ptr<Model>& P, MoveStats& Stats)
             if (log_verbose >=3) std::cerr<<"     Performing 2-way alignment\n";
             sample_alignment(*P.as<Parameters>(), b);
         }
-        sample_branch_length_(P,Stats,b);
+        if (t.has_branch_lengths() and t.can_set_branch_length(b)) sample_branch_length_(P,Stats,b);
 
         if (log_verbose >= 4)
         {
@@ -648,9 +730,10 @@ void realign_from_tips(owned_ptr<Model>& P, MoveStats& Stats)
             std::cerr<<"\n\n";
         }
 
-        three_way_topology_sample(P,Stats,b);
+        // We can't do this on a fixed topology.
+        // three_way_topology_sample(P,Stats,b);
     }
-    int AL1 = MCMC::alignment_length(*P.as<Parameters>());
+    int AL1 = alignment_length(*P.as<Parameters>());
     if (log_verbose>=4) std::cerr<<"realign_from_tips: |A1| = "<<AL1<<"\n";
 
     MCMC::Result result(2);
@@ -659,15 +742,17 @@ void realign_from_tips(owned_ptr<Model>& P, MoveStats& Stats)
     Stats.inc("realign_from_tips",result);
 }
 
-void walk_tree_sample_branch_lengths(owned_ptr<Model>& P, MoveStats& Stats) 
+void walk_tree_sample_branch_lengths(owned_ptr<context>& P, MoveStats& Stats) 
 {
+    if (log_verbose >= 3) std::cerr<<"\n\n[walk_tree_sample_branch_lengths]\n";
+
     Parameters& PP = *P.as<Parameters>();
-    vector<int> branches = walk_tree_path(PP.t(), PP[0].subst_root());
+    assert(PP.t().has_branch_lengths());
 
-    for(int i=0;i<branches.size();i++) 
+    vector<int> branches = walk_tree_path(PP.t(), PP.subst_root());
+
+    for(int b: branches)
     {
-        int b = branches[i];
-
         // Do a number of changes near branch @b
         sample_branch_length_(P,Stats,b);
     }
