@@ -21,9 +21,12 @@
 #include <iostream>
 #include <cmath>
 #include "util/assert.hh"
+#include "util/log-level.H"
+#include "util/settings.H"
 #include "sample.H"
 #include "probability/choose.H"
 #include "util/rng.H"
+#include "dp/4way.H"
 #include "dp/5way.H"
 #include "dp/alignment-sums.H"
 #include "alignment/alignment-util.H"
@@ -35,6 +38,7 @@
 using MCMC::MoveStats;
 using std::string;
 using std::vector;
+using std::optional;
 
 // We are sampling from a 5-way alignment (along 5 branches)
 
@@ -52,8 +56,6 @@ using std::vector;
 using std::abs;
 using std::valarray;
 
-using namespace A5;
-
 
 /// Update statistics counters for an NNI move.
 void NNI_inc(MoveStats& Stats, const string& name, MCMC::Result result,double L)
@@ -78,7 +80,7 @@ void NNI_inc(MoveStats& Stats, const string& name, MCMC::Result result,double L)
 	Stats.inc(name+"-2.0+", result);
 }
 
-// Do we need the different sample_two_nodes_base routines to use the same
+// Do we need the different sample_A5_base routines to use the same
 // sub-alignment ordering for different topologies?  No.
 //  o Sub-alignment order should affect only which paths are considered
 //  o We are essentially considering a set of paths for each topology
@@ -98,7 +100,7 @@ int two_way_topology_sample(vector<Parameters>& p,const vector<log_double_t>& rh
     orders[1] = A5::get_nodes_random(p[1].t(), b);
 
     try {
-	return sample_two_nodes_multi(p,orders,rho,true,false);
+	return sample_A5_multi(p,orders,rho);
     }
     catch (choose_exception<log_double_t>& c)
     {
@@ -110,39 +112,19 @@ int two_way_topology_sample(vector<Parameters>& p,const vector<log_double_t>& rh
 
 #include "slice-sampling.H"
 
-// Notes: the two-say slice sampler is more likely to accept topologies
-//        which have a low probability without branch adjustment.
-//        
-//        If we look at the distribution for accepted changes of
-//        log(P2)-log(P1) without adjusting the branch lengths,
-//        then 1% are below -3.23 for MH and -6.3 for slice.
+/*
+ * BROKEN: Tries to integrate out the alignment length along the internal branch on TWO topologies.
+           But never actually creates an internal node states for p[1].tree.
+ */
 
-//        Also, 2% are below -5.19 for slice, adjusting for the fact
-//        that there are fewer (half?) acceptances for slice.
-
-//        Finally, the bottom 5% for slice make up the bottom .7%
-//        (unadjusted) for slice...
-
-//        Experiment #1: load in a posterior sample, to find out what the acceptance
-//                       probability *should* be based only on the relative probabilities
-//                       of the topologies.
-//
-//        Experiment #2: for each proposal, do a long MCMC chain where we consider only two
-//                       topologies but allow branch lengths to vary, in order to find out
-//                       what the acceptance probability should be.
-
-// check the HMM type for a DIRECTED branch without going off the end of the array
-int HMM_type_for_branch(const Parameters& P, int b)
+void two_way_topology_slice_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
-    return P.branch_HMM_type(P.t().undirected(b));
-}
+    std::abort();
 
-void two_way_topology_slice_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
-{
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1) return;
+    if (log_verbose >= 3) std::cerr<<"[two_way_topology_slice_sample]\n";
 
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& nodes = order.nodes;
@@ -202,14 +184,14 @@ void two_way_topology_slice_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
     NNI_inc(Stats,"NNI (2-way,slice)", result, L);
 }
 
-void two_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void two_way_topology_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1) return;
+    if (log_verbose >= 3) std::cerr<<"[two_way_topology_sample]\n";
 
-    double slice_fraction = PP.load_value("NNI_slice_fraction",-0.25);
+    double slice_fraction = get_setting_or("NNI_slice_fraction",-0.25);
 
     if (not PP.variable_alignment() and uniform() < slice_fraction) {
 	two_way_topology_slice_sample(P,Stats,b);
@@ -274,7 +256,7 @@ void two_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
 
     int C = -1;
     try {
-	C = sample_two_nodes_multi(p,orders,rho,true,false);
+	C = sample_A5_multi(p,orders,rho);
     }
     catch (choose_exception<log_double_t>& c)
     {
@@ -290,25 +272,30 @@ void two_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
 
     MCMC::Result result(2);
 
+    double L0 = 1;
+    if (PP.t().has_branch_lengths())
+        L0 = PP.t().branch_length(b);
+
     result.totals[0] = (C>0)?1:0;
     // This gives us the average length of branches prior to successful swaps
     if (C>0)
-	result.totals[1] = p[0].t().branch_length(b);
+	result.totals[1] = L0;
     else
 	result.counts[1] = 0;
 
     if (smart)
-	NNI_inc(Stats,"NNI (2-way smart)", result, p[0].t().branch_length(b));
+	NNI_inc(Stats,"NNI (2-way smart)", result, L0);
     else
-	NNI_inc(Stats,"NNI (2-way stupid)", result, p[0].t().branch_length(b));
+	NNI_inc(Stats,"NNI (2-way stupid)", result, L0);
 }
 
-void two_way_NNI_SPR_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void two_way_NNI_SPR_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
+    if (not PP.t().has_branch_lengths()) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1) return;
+    if (log_verbose >= 3) std::cerr<<"[two_way_NNI_SPR_sample]\n";
 
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& nodes = order.nodes;
@@ -369,12 +356,13 @@ vector<int> NNI_branches(const TreeInterface& t, int b)
     return branches;
 }
 
-void two_way_NNI_and_branches_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void two_way_NNI_and_branches_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
+    if (not PP.t().has_branch_lengths()) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1) return;
+    if (log_verbose >= 3) std::cerr<<"[two_way_NNI_and_branches_sample]\n";
 
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& nodes = order.nodes;
@@ -402,8 +390,9 @@ void two_way_NNI_and_branches_sample(owned_ptr<Model>& P, MoveStats& Stats, int
 
 	auto factor = exp_to<log_double_t>(gaussian(0,0.05));
 
-	double L = p[1].t().branch_length( branches[i] ) * factor;
+	double L = p[1].t().branch_length( branches[i] ) * double(factor);
 
+        if (not p[1].t().can_set_branch_length(branches[i])) return;
 	p[1].setlength(branches[i], L);
 
 	ratio *= factor;
@@ -432,12 +421,12 @@ void two_way_NNI_and_branches_sample(owned_ptr<Model>& P, MoveStats& Stats, int
     NNI_inc(Stats,"NNI (2-way) + branches", result, p[0].t().branch_length(b));
 }
 
-void two_way_NNI_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void two_way_NNI_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1) return;
+    if (log_verbose >= 3) std::cerr<<"[two_way_NNI_sample]\n";
 
     double U = uniform();
     if (U < 0.33333333) {
@@ -449,13 +438,15 @@ void two_way_NNI_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
 	two_way_NNI_and_branches_sample(P,Stats,b);
 }
 
-void three_way_topology_sample_slice(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void three_way_topology_sample_slice(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
     if (PP.variable_alignment()) return;
 
+    if (log_verbose >= 3) std::cerr<<"[three_way_topology_sample_slice]\n";
+
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& nodes = order.nodes;
 
@@ -527,20 +518,59 @@ void three_way_topology_sample_slice(owned_ptr<Model>& P, MoveStats& Stats, int
     NNI_inc(Stats,"NNI (3-way,slice)", result, L);
 }
 
-void three_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+optional<log_double_t>& operator*=(optional<log_double_t>& pr1, const optional<log_double_t>& pr2)
+{
+    if (pr1)
+    {
+	if (pr2)
+	    pr1.value() *= pr2.value();
+	else
+	    pr1 = {};
+    }
+
+    return pr1;
+}
+
+optional<log_double_t> optimize(Parameters& P, const vector<int>& nodes)
+{
+    optional<log_double_t> ratio = 1;
+    if (uniform() < 0.5)
+    {
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[0]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[1]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[2]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[3]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[4]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[5]);
+    }
+    else
+    {
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[5]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[4]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[3]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[5], nodes[2]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[1]);
+	if (not ratio) return {};
+	ratio *= tri_sample_alignment_ratio(P, nodes[4], nodes[0]);
+    }
+    return ratio;
+}
+
+void two_way_topology_5A_sample(owned_ptr<context>& P, MoveStats& /*Stats*/, int b)
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1)
-	return;
-
-    double slice_fraction = PP.load_value("NNI_slice_fraction",-0.25);
-
-    if (not PP.variable_alignment() and uniform() < slice_fraction) {
-	three_way_topology_sample_slice(P,Stats,b);
-	return;
-    }
+    if (log_verbose >= 3) std::cerr<<"[two_way_topology_5A_sample]\n";
 
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& nodes = order.nodes;
@@ -549,12 +579,76 @@ void three_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
 
     int b1 = PP.t().find_branch(nodes[4],nodes[1]);
     int b2 = PP.t().find_branch(nodes[5],nodes[2]);
-    int b3 = PP.t().find_branch(nodes[5],nodes[3]);
+
+    //------ Generate Topologies and alter caches ------///
+
+    // 1. Resample alignments on original topology and compute reverse_sampling_pr/forward_sampling_pr
+    vector< A5::hmm_order > orders(2);
+    auto P0 = PP;
+    auto Like0A = P0.likelihood();
+    auto Prob0A = P0.probability();
+    auto PrAl0A = P0.prior_alignment();
+
+    orders[0] = A5::get_nodes_random(P0.t(), b);
+    optional<log_double_t> ratio = 1;
+    ratio *= optimize(P0, orders[0].nodes);
+    if (not ratio) return;
+
+    auto Like0B = P0.likelihood();
+    auto Prob0B = P0.probability();
+    auto PrAl0B = P0.prior_alignment();
+
+    // 2. Resample alignment on new topology and compute reverse_sampling_pr/forward_sampling_pr
+    auto P1 = P0;
+    P1.NNI_discard_alignment(b1, b2);
+    // Internal node states may be inconsistent after this: p[1].alignment_prior() undefined!
+    orders[1] = A5::get_nodes_random(P1.t(), b);
+
+    vector<Parameters> p({P0,P1});
+    auto ratio_tree = sample_A5_ratio(p, orders, {1,1});
+    ratio *= ratio_tree;
+    if (not ratio) return;
+
+    P1 = p[1];
+
+    bool accept1 = accept_MH(P0, P1, *ratio_tree);
+
+    auto Like1A = P1.likelihood();
+    auto Prob1A = P1.probability();
+    auto PrAl1A = P1.prior_alignment();
+
+    // 3. Resampling alignments on new topology and compute reverse_sampling_pr/forward_sampling_pr
+    ratio *= optimize(P1, orders[1].nodes);
+    if (not ratio) return;
+
+    auto Like1B = P1.likelihood();
+    auto Prob1B = P1.probability();
+    auto PrAl1B = P1.prior_alignment();
+
+    // 4. Accept/reject the new topology with resampled alignments given the proposal ratio.
+    bool accept2 = perform_MH(PP, P1, *ratio);
+    std::cerr<<std::boolalpha<<"[two_way_topology_5A_sample] accept1 = "<<accept1<<"        accept2 = "<<accept2<<"\n";
+    if (accept2 and not accept1)
+    {
+	std::cerr<<"Prob0A = "<<Prob0A <<" Like0A = "<<Like0A<<" PrAl0A = "<<PrAl0A<<"\n";
+	std::cerr<<"Prob0B = "<<Prob0B <<" Like0B = "<<Like0B<<" PrAl0B = "<<PrAl0B<<"\n";
+	std::cerr<<"Prob1A = "<<Prob1A <<" Like1A = "<<Like1A<<" PrAl1A = "<<PrAl1A<<"\n";
+	std::cerr<<"Prob1B = "<<Prob1B <<" Like1B = "<<Like1B<<" PrAl1B = "<<PrAl1B<<"\n";
+	std::cerr<<"\n\n";
+    }
+}
+
+void three_way_NNI_sample(Parameters& PP, MoveStats& Stats, int b, int b1, int b2, int b3)
+{
+    PP.select_root(b);
+    PP.cache_likelihood_branches();
 
     //------ Generate Topologies and alter caches ------///
     vector<Parameters> p(3,PP);
 
-    double L0 = PP.t().branch_length(b);
+    double L0 = 1;
+    if (PP.t().has_branch_lengths())
+        L0 = PP.t().branch_length(b);
 
     vector< A5::hmm_order > orders(3);
     orders[0] = A5::get_nodes_random(p[0].t(), b);
@@ -572,7 +666,7 @@ void three_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
     //------ Resample alignments and select topology -----//
 
     try {
-	int C = sample_two_nodes_multi(p,orders,rho,true,false);
+	int C = sample_A5_multi(p,orders,rho);
 
         if (C != -1) {
             PP = p[C];
@@ -596,13 +690,237 @@ void three_way_topology_sample(owned_ptr<Model>& P, MoveStats& Stats, int b)
     }
 }
 
-void three_way_topology_and_alignment_sample(owned_ptr<Model>& P, MoveStats& Stats, int b) 
+void three_way_NNI_A4_sample(Parameters& PP, MoveStats& Stats, int b, int b1, int b2, int b3)
+{
+    PP.select_root(b);
+    PP.cache_likelihood_branches();
+
+    //------ Generate Topologies and alter caches ------///
+    vector<Parameters> p(3,PP);
+
+    double L0 = 1;
+    if (PP.t().has_branch_lengths())
+        L0 = PP.t().branch_length(b);
+
+    vector< A4::hmm_order > orders(3);
+    orders[0] = A4::get_nodes_random(p[0].t(), b);
+
+    // Internal node states may be inconsistent after this: p[1].alignment_prior() undefined!
+    p[1].NNI_discard_alignment(b1, b2);
+    orders[1] = A4::get_nodes_random(p[1].t(), b);
+
+    // Internal node states may be inconsistent after this: p[2].alignment_prior() undefined!
+    p[2].NNI_discard_alignment(b1, b3);
+    orders[2] = A4::get_nodes_random(p[2].t(), b);
+
+    const vector<log_double_t> rho(3,1);
+
+    //------ Resample alignments and select topology -----//
+
+    try {
+        int C = sample_A4_multi(p,orders,rho);
+
+        if (C != -1) {
+            PP = p[C];
+        }
+
+        MCMC::Result result(2);
+
+        result.totals[0] = (C>0)?1:0;
+        // This gives us the average length of branches prior to successful swaps
+        if (C>0)
+            result.totals[1] = L0;
+        else
+            result.counts[1] = 0;
+
+        NNI_inc(Stats,"NNI (3-way)", result, L0);
+    }
+    catch (choose_exception<log_double_t>& c)
+    {
+        c.prepend(__PRETTY_FUNCTION__);
+        throw c;
+    }
+}
+
+void three_way_topology_sample(owned_ptr<context>& P, MoveStats& Stats, int b) 
 {
     Parameters& PP = *P.as<Parameters>();
     if (PP.t().is_leaf_branch(b)) return;
 
-    if (PP.variable_alignment() and HMM_type_for_branch(PP,b) == 1)
-	return;
+    if (log_verbose >= 3) std::cerr<<"[three_way_topology_sample]\n";
+
+    double slice_fraction = get_setting_or("NNI_slice_fraction",-0.25);
+
+    if (not PP.variable_alignment() and uniform() < slice_fraction)
+    {
+	three_way_topology_sample_slice(P, Stats, b);
+    }
+    else
+    {
+	A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
+	const auto& nodes = order.nodes;
+
+	int b1 = PP.t().find_branch(nodes[4],nodes[1]);
+	int b2 = PP.t().find_branch(nodes[5],nodes[2]);
+	int b3 = PP.t().find_branch(nodes[5],nodes[3]);
+
+        three_way_NNI_sample(PP, Stats, b, b1, b2, b3);
+    }
+}
+
+/* Like NNI, except
+ * + we can only swap branches pointing away from the root
+ * + we need to avoid children being older than their parents.
+ * + there may be a degree 2 node.
+ *
+ * Specificially, we need u to be younger than y.
+ *
+ *                       v
+ *                       |
+ *                    ---x---           ---x---
+ *                    |     |           |     |
+ *                  --y--   |    OR   --y--   |
+ *                  |   |   |         |   |   |
+ *                  z   w   u         z   w   u
+ *
+ * Here, tipward_branches[0] = (y,z)
+ *       tipward_branches[1] = (y,w)
+ *       other_branches[0]   = (x,u)
+ *       other_branches[1]   = (x,v) if x not the root
+ *       b           = (x,y)
+ */
+
+
+void three_way_time_tree_NNI_sample(owned_ptr<context>& P, MoveStats& Stats, int b)
+{
+    if (log_verbose >= 3) std::cerr<<"[three_way_time_tree_NNI_sample]\n";
+
+    Parameters& PP = *P.as<Parameters>();
+    auto T = PP.t();
+
+    // 1. Skip if this is not an internal branch
+    if (T.is_leaf_branch(b)) return;
+
+    // 2. Point branch away b = (x,y) away from root.
+    if (not T.away_from_root(b)) b = T.reverse(b);
+
+    int x = T.source(b);
+    int y = T.target(b);
+
+    assert(T.degree(x) == 2 or T.degree(x) == 3);
+    // We shouldn't be seeing degree-2 nodes except on the root branch.
+    assert(T.degree(y) == 3);
+
+    // 3. Get branches to child nodes z and w
+    auto tipward_branches = T.branches_after(b);
+    assert(tipward_branches.size() == 2);
+    int b2 = tipward_branches[0]; // (y,z)
+    int b3 = tipward_branches[1]; // (y,w)
+
+    // 4. Get branch to sibling node u
+    auto other_branches = T.branches_after(T.reverse(b));
+    assert(other_branches.size() == 1 or other_branches.size() == 2);
+    if (other_branches.size() == 2)
+    {
+	if (not T.away_from_root(other_branches[0]))
+	    std::swap(other_branches[0],other_branches[1]);
+	assert(not T.away_from_root(other_branches[1]));
+    }
+
+    int b1 = other_branches[0]; // (x,u)
+    int u = T.target(other_branches[0]);
+    assert(T.away_from_root(b1));
+
+    // 5. Check that node u is YOUNGER THAN node y
+    if (T.node_time(u) > T.node_time(y)) return;
+
+    if (other_branches.size() == 2)
+    {
+	// 6a. x is NOT the root.
+
+	// We consider interchanging b1 <-> b2 and b1 <-> b3
+	three_way_NNI_sample(PP, Stats, b, b1, b2, b3);
+    }
+    else
+    {
+	// 6b. x is the root.
+	three_way_NNI_A4_sample(PP, Stats, b, b1, b2, b3);
+
+	// We could also create moves to resample the alignment on 2 adjacent branches.
+    }
+}
+
+void three_way_topology_and_A5_2D_sample(owned_ptr<context>& P, MoveStats& Stats, int b, std::optional<int> bandwidth)
+{
+    Parameters& PP = *P.as<Parameters>();
+    auto t = PP.t();
+    if (t.is_leaf_branch(b)) return;
+
+    if (log_verbose >= 3) std::cerr<<"[three_way_topology_and_A5_2D_sample]\n";
+
+    //--------- Generate the Different Topologies -------//
+    // We do NOT always resample the connection between two_way_nodes [0] and [4].
+    // Should we?
+    PP.select_root(b);
+    PP.cache_likelihood_branches();
+
+    //------ Generate Topologies and alter caches ------///
+    vector<Parameters> p(3,PP);
+
+    double L0 = 1;
+    if (PP.t().has_branch_lengths())
+	L0 = PP.t().branch_length(b);
+
+    vector< A5::hmm_order > orders(3);
+    orders[0] = A5::get_nodes_random(PP.t(), b);
+
+    auto& nodes = orders[0].nodes;
+    int b1 = t.find_branch(nodes[4], nodes[1]);
+    int b2 = t.find_branch(nodes[5], nodes[2]);
+    int b3 = t.find_branch(nodes[5], nodes[3]);
+
+    // Internal node states may be inconsistent after this: p[1].alignment_prior() undefined!
+    p[1].NNI_discard_alignment(b1, b2);
+    orders[1] = A5::get_nodes_random(p[1].t(), b);
+
+    // Internal node states may be inconsistent after this: p[2].alignment_prior() undefined!
+    p[2].NNI_discard_alignment(b1, b3);
+    orders[2] = A5::get_nodes_random(p[2].t(), b);
+
+    const vector<log_double_t> rho(3,1);
+
+    int C = -1;
+    try {
+	C = sample_A5_2D_multi(p, orders, rho, bandwidth);
+    }
+    catch (choose_exception<log_double_t>& c)
+    {
+	c.prepend(__PRETTY_FUNCTION__);
+	throw c;
+    }
+
+    if (C != -1) {
+	PP = p[C];
+    }
+
+    MCMC::Result result(2);
+
+    result.totals[0] = (C>0)?1:0;
+    // This gives us the average length of branches prior to successful swaps
+    if (C>0)
+	result.totals[1] = L0;
+    else
+	result.counts[1] = 0;
+
+    NNI_inc(Stats,"NNI (5-way) + A", result, L0);
+}
+
+void three_way_topology_and_A3_2D_sample(owned_ptr<context>& P, MoveStats& Stats, int b, std::optional<int> bandwidth) 
+{
+    Parameters& PP = *P.as<Parameters>();
+    if (PP.t().is_leaf_branch(b)) return;
+
+    if (log_verbose >= 3) std::cerr<<"[three_way_topology_and_A3_2D_sample]\n";
 
     A5::hmm_order order = A5::get_nodes_random(PP.t(), b);
     const auto& two_way_nodes = order.nodes;
@@ -614,10 +932,11 @@ void three_way_topology_and_alignment_sample(owned_ptr<Model>& P, MoveStats& Sta
 
     PP.select_root(b);
     PP.cache_likelihood_branches();
+    auto t = PP.t();
 
-    int b1 = PP.t().find_branch(two_way_nodes[4],two_way_nodes[1]);
-    int b2 = PP.t().find_branch(two_way_nodes[5],two_way_nodes[2]);
-    int b3 = PP.t().find_branch(two_way_nodes[5],two_way_nodes[3]);
+    int b1 = t.find_branch(two_way_nodes[4],two_way_nodes[1]);
+    int b2 = t.find_branch(two_way_nodes[5],two_way_nodes[2]);
+    int b3 = t.find_branch(two_way_nodes[5],two_way_nodes[3]);
 
     vector<Parameters> p(3,PP);
 
@@ -635,7 +954,7 @@ void three_way_topology_and_alignment_sample(owned_ptr<Model>& P, MoveStats& Sta
 
     int C = -1;
     try {
-	C = sample_tri_multi(p,nodes,rho,true,true);
+	C = sample_tri_multi(p,nodes,rho);
     }
     catch (choose_exception<log_double_t>& c)
     {
@@ -658,3 +977,13 @@ void three_way_topology_and_alignment_sample(owned_ptr<Model>& P, MoveStats& Sta
   
     NNI_inc(Stats,"NNI (3-way) + A", result, L0);
 }
+
+void three_way_topology_and_alignment_sample(owned_ptr<context>& P, MoveStats& Stats, int b, std::optional<int> bandwidth)
+{
+    double NNI_A5_fraction = get_setting_or("A5_fraction",0.0);
+    if (uniform() < NNI_A5_fraction)
+        three_way_topology_and_A5_2D_sample(P, Stats, b, bandwidth);
+    else
+        three_way_topology_and_A3_2D_sample(P, Stats, b, bandwidth);
+}
+
