@@ -23,7 +23,9 @@
 /// \brief This file implements classes and functions for uniform slice sampling.
 ///
 
+#include <tuple>
 #include "util/assert.hh"
+#include "util/log-level.H"
 #include "slice-sampling.H"
 #include "util/rng.H"
 #include "probability/choose.H"
@@ -63,9 +65,12 @@ double context_slice_function::operator()(double x)
     // How can we automate this so that it is called only once?
     auto ratio = C.heated_probability_ratios(C0);
     if (ratio.variables_changed)
+    {
         throw variables_changed_exception("Variable changed during slice sampling!");
-
-    current_fn_value = ratio.total_ratio();
+        current_fn_value = 0;
+    }
+    else
+        current_fn_value = ratio.total_ratio();
 
     return operator()();
 }
@@ -93,7 +98,7 @@ double random_variable_slice_function::current_value() const
 random_variable_slice_function::random_variable_slice_function(context_ref& c, const bounds<double>& bounds, int rv)
     :context_slice_function(c, bounds)
 {
-    if (auto m = C.get_modifiable_reg(rv))
+    if (auto m = C.find_modifiable_reg(rv))
 	r_mod = *m;
     else
 	throw myexception()<<"No modifiable reg for slice function!";
@@ -123,7 +128,7 @@ double integer_random_variable_slice_function::current_value() const
 integer_random_variable_slice_function::integer_random_variable_slice_function(context_ref& c, const bounds<int>& bounds, int rv)
     :context_slice_function(c, convert_bounds(bounds))
 {
-    if (auto m = C.get_modifiable_reg(rv))
+    if (auto m = C.find_modifiable_reg(rv))
 	r_mod = *m;
     else
 	throw myexception()<<"No modifiable reg for slice function!";
@@ -138,7 +143,7 @@ void branch_length_slice_function::set_value(double l)
 
 double branch_length_slice_function::current_value() const
 {
-    return static_cast<Parameters&>(C).t().branch_length(b);
+    return static_cast<const Parameters&>(C).t().branch_length(b);
 }
 
 branch_length_slice_function::branch_length_slice_function(Parameters& P,int b_)
@@ -147,6 +152,25 @@ branch_length_slice_function::branch_length_slice_function(Parameters& P,int b_)
     set_lower_bound(0);
 }
 
+// ******************************* node time length slice function *************************************** //
+
+void node_time_slice_function::set_value(double t)
+{
+    static_cast<Parameters&>(C).t().set_node_time(n,t);
+}
+
+double node_time_slice_function::current_value() const
+{
+    return static_cast<Parameters&>(C).t().node_time(n);
+}
+
+node_time_slice_function::node_time_slice_function(Parameters& P,int n_)
+    :context_slice_function(P),n(n_)
+{
+}
+
+// ******************************* branch length slice function *************************************** //
+
 double alignment_branch_length_slice_function::operator()(double x)
 {
     count++;
@@ -232,60 +256,60 @@ slide_node_slice_function::slide_node_slice_function(Parameters& P,int i1,int i2
  * Joint scaling of branch lengths and scales so that the T*R=D remains constant.
  */
 
-void scale_means_only_slice_function::set_value(double t)
+void scale_groups_slice_function::set_value(double t)
 {
     log_current_factor = t;
     double scale = exp(log_current_factor);
 
-    auto& P = static_cast<Parameters&>(C);
-
     // Scale the branch-scaling factor for each partition.
     for(int i=0; i<initial_scales.size(); i++)
-	P.set_branch_scale(i, initial_scales[i]*scale);
+	C.set_modifiable_value(r_scales[i], initial_scales[i]*scale);
 
     // Scale the branch lengths in the opposite direction
     for(int b=0; b<initial_branch_lengths.size(); b++)
-	P.setlength_unsafe(b, initial_branch_lengths[b]/scale);
+	C.set_modifiable_value(r_branch_lengths[b], initial_branch_lengths[b]/scale);
 }
 
-double scale_means_only_slice_function::log_average_scale() const
+double scale_groups_slice_function::log_average_scale() const
 {
     return log(initial_average_scale) + log_current_factor;
 }
 
-double scale_means_only_slice_function::operator()()
+double scale_groups_slice_function::operator()()
 {
     // return pi * (\sum_i \mu_i)^(n-B)
     return context_slice_function::operator()() + log_average_scale()*((int)initial_scales.size() - (int)initial_branch_lengths.size());
 }
 
-double scale_means_only_slice_function::current_value() const
+double scale_groups_slice_function::current_value() const
 {
     return log_current_factor;
 }
 
-scale_means_only_slice_function::scale_means_only_slice_function(Parameters& P)
-    :context_slice_function(P)
+scale_groups_slice_function::scale_groups_slice_function(context_ref& C, const std::vector<int>& ss, const std::vector<int>& ls)
+    :context_slice_function(C),
+     r_scales(ss),
+     r_branch_lengths(ls)
 { 
     // 1. Set up initial scales
-    if (P.n_branch_scales() == 0)
+    if (n_scales() == 0)
         throw myexception()<<"Can't do scale_means_only_slice function if there are no scales!";
 
     initial_average_scale = 0;
-    initial_scales.resize(P.n_branch_scales());
+    initial_scales.resize(n_scales());
     for(int i=0;i<initial_scales.size();i++)
     {
-        initial_scales[i] = P.get_branch_scale(i);
-        initial_average_scale += initial_scales[i];
+        initial_scales[i] = C.get_modifiable_value(r_scales[i]).as_double();
+        initial_average_scale += std::abs(initial_scales[i]);
     }
     initial_average_scale /= initial_scales.size();
 
     // FIXME: We should be able to assert that all of the scales are modifiable.
 
     // 3. Set up initial branch_lengths
-    initial_branch_lengths.resize(P.t().n_branches());
+    initial_branch_lengths.resize(n_branch_lengths());
     for(int b=0; b<initial_branch_lengths.size(); b++)
-        initial_branch_lengths[b] = P.t().branch_length(b);
+        initial_branch_lengths[b] = C.get_modifiable_value(r_branch_lengths[b]).as_double();
 
     // FIXME: We should be able to assert that all of the branch lengths are modifiable.
 
@@ -297,11 +321,12 @@ scale_means_only_slice_function::scale_means_only_slice_function(Parameters& P)
     //     log(initial_average_scale) + log_current_factor \in [-40,40]
     //                                  log_current_factor \in [-40 - log(initial_average_scale), 40 - log(initial_average_scale)]
     double shift = -log(initial_average_scale);
+    assert(std::isfinite(shift));
 
     b = between<double>(-40+shift,40+shift);
 
 #ifndef NDEBUG
-    std::clog<<"bounds on t are "<<b<<std::endl;
+    if (log_verbose >= 4) std::clog<<"bounds on t are "<<b<<std::endl;
 #endif
 }
 
@@ -372,18 +397,21 @@ constant_sum_modifiable_slice_function::constant_sum_modifiable_slice_function(c
 std::pair<double,double> 
 find_slice_boundaries_stepping_out(double x0,slice_function& g,double logy, double w,int m)
 {
+    assert(x0 + w > x0);
     assert(g.in_range(x0));
 
     double u = uniform()*w;
     double L = x0 - u;
-    double R = x0 + (w-u);
+    double R = L + w;
+    assert(L < x0);
+    assert(x0 < R);
 
     // Expand the interval until its ends are outside the slice, or until
     // the limit on steps is reached.
 
     //  std::cerr<<"!!    L0 = "<<L<<"   x0 = "<<x0<<"   R0 = "<<R<<"\n";
     if (m>1) {
-	int J = binomial(m, 0.5);
+	int J = uniform_int(0,m);
 	int K = m-J;
 
 	while (J>0 and (not g.below_lower_bound(L)) and g(L)>logy) {
@@ -408,23 +436,92 @@ find_slice_boundaries_stepping_out(double x0,slice_function& g,double logy, doub
 	    R += w;
     }
 
-    // Shrink interval to lower and upper bounds.
-
-    if (g.below_lower_bound(L)) L = *g.lower_bound;
-    if (g.above_upper_bound(R)) R = *g.upper_bound;
 
     assert(L < R);
 
     //  std::cerr<<"[]    L0 = "<<L<<"   x0 = "<<x0<<"   R0 = "<<R<<"\n";
 
-    return std::pair<double,double>(L,R);
+    return {L,R};
+}
+
+std::tuple<double,double,std::optional<double>,std::optional<double>>
+find_slice_boundaries_doubling(double x0,slice_function& g,double logy, double w, int K)
+{
+    assert(x0 + w > x0);
+    assert(g.in_range(x0));
+
+    double u = uniform()*w;
+    double L = x0 - u;
+    double R = L + w;
+    assert(L < x0);
+    assert(x0 < R);
+
+    std::optional<double> gL_cached;
+    auto gL = [&]() {
+        if (not gL_cached)
+            gL_cached = (g.in_range(L))?g(L):log_0;
+        return *gL_cached;
+    };
+
+    std::optional<double> gR_cached;
+    auto gR = [&]() {
+        if (not gR_cached)
+            gR_cached = (g.in_range(R))?g(R):log_0;
+        return *gR_cached;
+    };
+
+    auto too_large = [](double L, double R, double w)
+    {
+        double M = (L+R)/2;
+        double W = R-L;
+        assert(W > 0);
+        bool ok = (L < M) and (M < R) and (W+w > W) and (L-w < L) and (R+w>R);
+        return not ok;
+    };
+
+    while ( K > 0 and (gL() > logy or gR() > logy))
+    {
+        if (log_verbose >= 4)
+            std::cerr<<"!!    L0 = "<<L<<" (g(L) = "<<gL()<<")  x0 = "<<x0<<"   R0 = "<<R<<" (g(R) = "<<gR()<<")\n";
+
+        double W2 = (R-L);
+        if (uniform() < 0.5)
+        {
+            double L2 = L - W2;
+            if (too_large(L2, R, w))
+                break;
+            L = L2;
+            gL_cached = {};
+        }
+        else
+        {
+            double R2 = R + W2;
+            if (too_large(L, R2, w))
+                break;
+            R = R2;
+            gR_cached = {};
+        }
+
+        K --;
+    }
+
+    assert(L < R);
+    assert( L < (L+R)/2 and (L+R)/2 < R);
+
+    //  std::cerr<<"[]    L0 = "<<L<<"   x0 = "<<x0<<"   R0 = "<<R<<"\n";
+
+    return {L,R,gL_cached,gR_cached};
 }
 
 // Does this x0 really need to be the original point?
 // I think it just serves to let you know which way the interval gets shrunk...
 
-double search_interval(double x0,double& L, double& R, slice_function& g,double logy)
+double search_interval(double x0,double L, double R, slice_function& g, double logy)
 {
+    // Shrink interval to lower and upper bounds.
+    if (g.below_lower_bound(L)) L = *g.lower_bound;
+    if (g.above_upper_bound(R)) R = *g.upper_bound;
+
     //  assert(g(x0) > g(L) and g(x0) > g(R));
     assert(g(x0) >= logy);
     assert(L < R);
@@ -449,6 +546,7 @@ double search_interval(double x0,double& L, double& R, slice_function& g,double
 	else
 	    L = x1;
     }
+    std::cerr.precision(17);
     std::cerr<<"Warning!  Is size of the interval really ZERO?"<<std::endl;
     double logy_x0 = g(x0);  
     std::cerr<<"    L0 = "<<L0<<"   x0 = "<<x0<<"   R0 = "<<R0<<std::endl;
@@ -460,49 +558,136 @@ double search_interval(double x0,double& L, double& R, slice_function& g,double
     return x0;
 }
 
-double slice_sample_(double x0, slice_function& g, double w, int m)
+bool pre_slice_sampling_check_OK(double x0, slice_function& g)
 {
     // If x is not in the range then this could be a range that is reduced to avoid loss of precision.
     if (not g.in_range(x0))
     {
         if (log_verbose >= 4) std::cerr<<x0<<" not in range!";
-        return x0;
+        return false;
     }
 
     assert(g.in_range(x0));
 
     double gx0 = g();
+    double gx0_v2 = gx0;
     if (log_verbose >= 4)
+        gx0_v2 = g(x0);
+#ifndef NDEBUG
+    else
+        gx0_v2 = g(x0);
+#endif
+    if (std::abs(gx0 - gx0_v2) > 1.0e-9)
+        throw myexception()<<"Error: slice_sampling: g() = "<<gx0<<"   g(x0) = "<<gx0_v2<<"   diff = "<<std::abs(gx0 - gx0_v2);
+
+    return true;
+}
+
+bool can_propose_same_interval_doubling(double x0, double x1, double w, double L, double R, std::optional<double> gL_cached, std::optional<double> gR_cached, slice_function& g, double log_y)
+{
+    bool D = false;
+
+    auto gL = [&]() {
+        if (not gL_cached)
+            gL_cached = (g.in_range(L))?g(L):log_0;
+        return *gL_cached;
+    };
+
+    auto gR = [&]() {
+        if (not gR_cached)
+            gR_cached = (g.in_range(R))?g(R):log_0;
+        return *gR_cached;
+    };
+
+    bool ok = true;
+    while (ok and R-L > 1.1*w)
     {
-	std::cerr<<" gx0 = "<<gx0<<std::endl;
-	std::cerr<<" g(x0) = "<<g(x0)<<std::endl;
+        double M = (R+L)/2;
+        assert( L < M and M < R);
+
+        // Check if x0 and x1 are in different halves of the interval.
+        if ((x0 < M and x1 >= M) or (x0 >= M and x1 < M))
+            D = true;
+
+        if (x1 < M)
+        {
+            R = M;
+            gR_cached = {};
+        }
+        else
+        {
+            L = M;
+            gL_cached = {};
+        }
+
+        if (D and log_y >= gL() and log_y >= gR())
+            ok = false;
     }
 
-#ifndef NDEBUG
-    volatile double diff = gx0 - g(x0);
-    assert(std::abs(diff) < 1.0e-9);
-#endif
+    // FIXME - this is clunky.  Do we really want to set x by evaluate g( )?
+    if (D)
+    {
+        // We may have set x to L or R, so reset it to the right values.
+        if (ok)
+            g(x1);
+        else
+            g(x0);
+    }
 
-    // Determine the slice level, in log terms.
+    return ok;
+}
 
-    double logy = gx0 - exponential(1);
+double slice_sample_stepping_out_(double x0, slice_function& g, double w, int m)
+{
+    // 0. Check that the values are OK
+    if (not pre_slice_sampling_check_OK(x0, g))
+        return x0;
 
-    // Find the initial interval to sample from.
+    // 1. Determine the slice level, in log terms.
+    double logy = g() - exponential(1);
 
-    std::pair<double,double> interval = find_slice_boundaries_stepping_out(x0,g,logy,w,m);
-    double L = interval.first;
-    double R = interval.second;
-
-    // Sample from the interval, shrinking it on each rejection
+    // 2. Find the initial interval to sample from.
+    auto [L,R] = find_slice_boundaries_stepping_out(x0,g,logy,w,m);
 
+    // 3. Sample from the interval, shrinking it on each rejection
     return search_interval(x0,L,R,g,logy);
 }
 
+// We need to SET the value INSIDE this routine.
+// Are we assuming that calling g sets the value?
+double slice_sample_doubling_(double x0, slice_function& g, double w, int m)
+{
+    if (log_verbose >= 4)
+    {
+        std::cerr<<"slice_sampling_doubling_: x0 = "<<x0<<" w = "<<w<<"\n";
+    }
+
+    // 0. Check that the values are OK
+    if (not pre_slice_sampling_check_OK(x0, g))
+        return x0;
+
+    // 1. Determine the slice level, in log terms.
+    double logy = g() - exponential(1);
+
+    // 2. Find the initial interval to sample from.
+    auto [L,R,gL_cached,gR_cached] = find_slice_boundaries_doubling(x0,g,logy,w,m);
+
+    // 3. Sample from the interval, shrinking it on each rejection
+    double x1 = search_interval(x0,L,R,g,logy);
+
+    // 4. Check that we can propose the same interval from x2
+    // We need to SET the value INSIDE this routine if we recompute g().
+    if (can_propose_same_interval_doubling(x0, x1, w, L, R, gL_cached, gR_cached, g, logy))
+        return x1;
+    else
+        return x0;
+}
+
 double slice_sample(double x0, slice_function& g,double w, int m)
 {
     try
     {
-        return slice_sample_(x0, g, w, m);
+        return slice_sample_doubling_(x0, g, w, m);
     }
     catch (variables_changed_exception& e)
     {
@@ -687,7 +872,7 @@ std::pair<int,double> search_multi_intervals(vector<double>& X0,
 	assert(intervals[C].first <= X0[C] and X0[C] < intervals[C].second);
     }
 }
-					   
+
 std::pair<int,double> slice_sample_multi(vector<double>& X0, vector<slice_function*>& g, double w, int m)
 {
     int N = g.size();
