@@ -2,12 +2,12 @@
 #include <vector>
 #include <list>
 #include "models/parse.H"
-#include "models/model.H"
 #include "util/myexception.H"
 #include "util/string/join.H"
 #include "util/range.H" // for reverse( )
 #include "rules.H"
-#include "setup.H"
+#include "compile.H"
+#include "models/driver.hh"
 
 using std::vector;
 using std::list;
@@ -15,53 +15,6 @@ using std::string;
 using std::pair;
 using std::optional;
 
-optional<pair<string,string>> split_keyword(const string& s, char c)
-{
-    for(int i=0;i<s.size();i++)
-    {
-	if (s[i] == '[' or s[i] == ']') return {};
-	if (s[i] == c)
-	    return pair<string,string>({s.substr(0,i),s.substr(i+1)});
-    }
-    return {};
-}
-
-/// \brief Turn an expression of the form h1[a]+h2[b] -> {h1[a],h2[b]}
-///
-/// \param sstack A stack of strings that represent a substitution model.
-/// \param s The model name to match.
-/// \param args The possible argument.
-///
-vector<string> split_top_plus(const string& s)
-{
-    vector<string> terms;
-
-    // Split on '+' where depth == 0.
-    int pos = 0;
-    int depth = 0;
-    for(int i=0;i<s.size();i++)
-    {
-	if (s[i] == '[')
-	    depth++;
-	else if (s[i] == ']')
-	{
-	    depth--;
-	    if (depth < 0) throw myexception()<<"Too many ']' in string '"<<s<<"'";
-	}
-	else if (depth == 0 and s[i] == '+')
-	{
-	    terms.push_back(s.substr(pos,i-pos));
-	    pos = i+1;
-	}
-    }
-
-    if (depth != 0)
-	throw myexception()<<"Too many '[' in string '"<<s<<"'";
-  
-    terms.push_back(s.substr(pos,int(s.size())-pos));
-
-    return terms;
-}
 
 // Turn an expression of the form head[arg1, arg2, ..., argn] -> {head, arg1, arg2, ..., argn}.
 vector<string> split_args(string s)
@@ -92,7 +45,9 @@ vector<string> split_args(string s)
 	if ((s[i] == ']' or s[i] == ',') and depth == 1)
 	{
 	    assert(i >= start);
-	    args.push_back(s.substr(start,i-start));
+            auto arg = s.substr(start,i-start);
+            if (not arg.empty() or s[i] == ',')
+                args.push_back(arg);
 	    start = i+1;
 	}
 
@@ -111,25 +66,7 @@ vector<string> split_args(string s)
 
 ptree parse_type(const string& s)
 {
-    // 1. Split the head and the arguments
-    auto args = split_args(s);
-
-    // 2. Set the head
-    string head = args.front();
-    args.erase(args.begin());
-
-    ptree result;
-    result.put_value(head);
-
-    // 3. Set the arguments
-    for(const auto& arg: args)
-    {
-	if (arg.empty()) throw myexception()<<"Type '"<<s<<"' has empty argument";
-
-	result.push_back({"", parse_type(arg)});
-    }
-
-    return result;
+    return parse_type(s,"type");
 }
 
 ptree add_sample(const ptree& p)
@@ -139,108 +76,13 @@ ptree add_sample(const ptree& p)
     return p2;
 }
 
-ptree parse(const string& s);
-
-/// Parse strings of the form {~}head[value1, value2, key3=value3, ...]
-ptree parse_no_submodel(const string& s)
+ptree add_submodel(ptree term, const ptree& submodel)
 {
-    if (s.empty())
-	throw myexception()<<"term is empty!";
+    if (term.count("submodel"))
+        throw myexception()<<"Trying to specify a submodel with '+' when submodel already specified by keyword!";
 
-    // 0. Handle ~
-    if (not s.empty() and s[0] == '~')
-	return add_sample(parse_no_submodel(s.substr(1)));
-
-    // 1. Split the head and the arguments
-    auto args = split_args(s);
-
-    // 2. Set the head
-    string head = args.front();
-    args.erase(args.begin());
-
-    ptree result;
-    if (auto i = can_be_converted_to<int>(head))
-	result.put_value(*i);
-    else if (auto d = can_be_converted_to<double>(head))
-	result.put_value(*d);
-    else if (auto b = can_be_converted_to<bool>(head))
-	result.put_value(*b);
-    // Currently we leave literal (i.e. quoted) strings as strings w/ quotes.
-    else
-	result.put_value(head);
-
-    if (args.size() and not result.is_a<string>())
-	throw myexception()<<"Error parsing '"<<s<<"': constant '"<<head<<"' should not have arguments, but has "<<args.size()<<"!";
-
-    // 3. Attempt to set the supplied arguments
-    for(int i=0;i<args.size();i++)
-    {
-	pair<string,string> key_value;
-
-	// 3.1. 'key=value'
-	if (auto arg = split_keyword(args[i],'='))
-	{
-	    key_value = *arg;
-
-	    if (arg->first.empty())
-		throw myexception()<<"Parameter name missing in argument '"<<args[i]<<"'";
-	}
-	// 3.2 'key~value'
-	else if ((arg = split_keyword(args[i],'~')) and arg->first.size())
-	{
-	    key_value = *arg;
-	    key_value.second = "~"+key_value.second;
-	}
-	// 3.3. 'value'
-	else
-	    key_value.second = args[i];
-
-	if (key_value.first.empty() and key_value.second.empty())
-	    result.push_back({{},{}});
-	else
-	    result.push_back({key_value.first, parse(key_value.second)});
-    }
-
-    return result;
-}
-
-ptree parse(const string& s)
-{
-    // 1. Get the last head[args]
-    vector<string> terms = split_top_plus(s);
-
-    // 2. Parse the terms and handle submodels;
-    ptree expression;
-    for(int i=0;i<terms.size();i++)
-    {
-	auto& term = terms[i];
-	try
-	{
-	    ptree parsed_term = parse_no_submodel(term);
-	    if (parsed_term.is_null())        throw myexception()<<"Term is empty!";
-
-	    if (not expression.is_null())
-	    {
-		if (parsed_term.count("submodel"))    throw myexception()<<"Trying to specify a submodel with '+' when submodel already specified by keyword!";
-		parsed_term.push_back({"submodel", expression});
-	    }
-	    expression = parsed_term;
-
-	    assert(not expression.is_null());
-	}
-	catch (myexception& e)
-	{
-	    string eterm = term;
-	    if (i-1 >= 0)
-		eterm = "+"+eterm;
-	    if (i+1 < terms.size())
-		eterm = eterm + "+";
-	    e.prepend("Parsing '"+eterm+"': ");
-	    throw;
-	}
-    }
-
-    return expression;
+    term.push_back({"submodel", submodel});
+    return term;
 }
 
 void ptree_map(std::function<void(ptree&)> f, ptree& p)
@@ -344,17 +186,9 @@ void handle_positional_args(ptree& model, const Rules& R)
 
     auto head = model.get_value<string>();
 
-    if (head == "let")
-    {
-	if (model.size() != 2)
-	    throw myexception()<<"let: got "<<model.size()<<" arguments, 2 arguments required.\n  "<<model.show();
-	if (model[0].first.empty())
-	    throw myexception()<<"let: first argument must have variable name!\n  "<<model.show();
-	if (not model[1].first.empty())
-	    throw myexception()<<"let: second argument must not have a variable name!\n  "<<model.show();
-	model[1].first = "body";
-	return;
-    }
+    if (head == "!let") return;
+
+    if (head == "!Decls") return;
 
     if (is_function(model)) return;
 
@@ -381,6 +215,14 @@ void handle_positional_args(ptree& model, const Rules& R)
         return;
     }
 
+    if (not R.get_rule_for_func(head))
+    {
+	for(auto& [arg_name,value]: model)
+	    if (not arg_name.empty())
+		throw myexception()<<"Named argument '"<<arg_name<<"' not allowed for function '"<<head<<"' with no rule.";
+	return;
+    }
+
     auto rule = R.require_rule_for_func(head);
 
     int i=0;
@@ -424,10 +266,10 @@ void handle_positional_args(ptree& model, const Rules& R)
 }
 
 // Parse strings of the form head[args] + head[args] + ... + head[args]
-ptree parse(const Rules& R, const string& s)
+ptree parse(const Rules& R, const string& s, const string& what)
 {
     // 1. Get the last head[args]
-    auto model = parse(s);
+    auto model = parse_expression(s, what);
 
     // 2. Fill in keywords where they are not given
     auto f2 = [&](ptree& p) {handle_positional_args(p,R);};
@@ -436,6 +278,59 @@ ptree parse(const Rules& R, const string& s)
     return model;
 }
 
+ptree parse_defs(const Rules& R, const string& s)
+{
+    // 1. Parse the decls
+    auto defs = parse_defs(s, "declarations");
+
+    // 2. Fill in default arguments
+    auto f2 = [&](ptree& p) {handle_positional_args(p,R);};
+    for(auto& [var,value]: defs)
+	ptree_map(f2, value);
+
+    return defs;
+}
+
+optional<ptree> peel_sample(ptree p)
+{
+    if (p.has_value<string>() and p.get_value<string>() == "sample")
+	return p[0].second;
+    else if (p.has_value<string>() and p.get_value<string>() == "!let")
+    {
+        if (auto new_body = peel_sample(p[1].second))
+        {
+            p[1].second = *new_body;
+            return p;
+        }
+    }
+
+    return {};
+}
+
+optional<ptree> peel_sample_annotated(ptree p)
+{
+    auto& v = p.get_child("value");
+    if (v.has_value<string>() and v.get_value<string>() == "sample")
+	return v[0].second;
+    else if (v.has_value<string>() and v.get_value<string>() == "!let")
+    {
+        if (auto new_body = peel_sample_annotated(v[1].second))
+        {
+            v[1].second = *new_body;
+            return p;
+        }
+    }
+
+    return {};
+}
+
+bool is_operator(const string& s)
+{
+    return (s == "+" or s == "-" or s == "*" or s == "/");
+}
+
+// To properly parenthesize infix expressions, we could ... pass in a context argument
+// that says when they are the left or right argument of an infix operator?
 
 string unparse(const ptree& p)
 {
@@ -460,50 +355,77 @@ string unparse(const ptree& p)
 
     assert(p.has_value<string>());
     string s = p.get_value<string>();
-    if (s == "let")
+    if (s == "!let")
     {
 	string name = p[0].first;
-	return "let["+name+"="+unparse(p[0].second)+","+unparse(p[1].second)+"]";
+	return name+show_model(p[0].second)+";"+unparse(p[1].second);
     }
     else if (s == "function")
     {
-	string name = p[0].first;
-	return "function["+name+","+unparse(p[1].second)+"]";
+	string name = unparse(p[0].second);
+	return "|"+name+":"+unparse(p[1].second)+"|";
+    }
+    else if (s == "List")
+    {
+        vector<string> items;
+        for(auto& [_,item]: p)
+            items.push_back(unparse(item));
+        return "[" + join(items,", ") + "]";
+    }
+    else if (s == "Tuple")
+    {
+        vector<string> items;
+        for(auto& [_,item]: p)
+            items.push_back(unparse(item));
+        return "(" + join(items,", ") + ")";
     }
     else if (s == "sample")
 	return "~" + unparse(p.begin()->second);
+    else if (s == "negate")
+	return "-" + unparse(p.begin()->second);
+    else if (is_operator(s) and p.size() == 2)
+    {
+	// sometimes we might need parenthesis, right?
+        return unparse(p[0].second) + s + unparse(p[1].second);
+    }
 
-    if (s == "intToDouble")
+    else if (s == "intToDouble")
 	return unparse(p.get_child("x"));
-    if (s == "unit_mixture" or s == "multiMixtureModel")
+    else if (s == "unit_mixture" or s == "multiMixtureModel")
 	if (auto child = p.get_child_optional("submodel"))
 	    return unparse(*child);
     vector<string> args;
     optional<string> submodel;
-    for(const auto& pair: p)
+    bool positional = true;
+    int pos = 0;
+    for(const auto& [arg_name, arg]: p)
     {
-	if (pair.second.is_null())
-	{
-	    args.push_back("");
-	    continue;
-	}
+	if (arg.is_null())
+            positional = false;
+
 	// Don't print submodel arguments: move out to submodel + <this>
-	else if (pair.first == "submodel")
+	else if (arg_name == "submodel" and pos == 0)
 	{
+            positional = false;
 	    assert(not submodel);
-	    submodel = unparse(pair.second);
-	    args.push_back("");
+	    submodel = unparse(arg);
 	}
-	else
-	    args.push_back( unparse(pair.second) );
-	// FIXME: don't print get_state[state_name] if its a default value?
+	else if (positional)
+	    args.push_back( unparse(arg) );
+        else if (auto arg2 = peel_sample(arg))
+            args.push_back( arg_name + "~" + unparse(*arg2));
+        else
+            args.push_back( arg_name + "=" + unparse(arg));
+
+	pos++;
+	// With annotations, we don't print get_state[state_name] if its a default value.
     }
     while (args.size() and args.back() == "")
 	args.pop_back();
     if (not args.empty())
-	s = s + "[" + join(args,',') + "]";
+	s = s + "(" + join(args,", ") + ")";
     if (submodel)
-	s = *submodel + "+" + s;
+	s = *submodel + " +> " + s;
     return s;
 }
 
@@ -536,89 +458,125 @@ string unparse_annotated(const ptree& ann)
 	return p.get_value<string>();
 
     string s = p.get_value<string>();
-    if (s == "let")
+    if (s == "!let")
     {
-	string name = p[0].first;
-	return "let["+name+show_model_annotated(p[0].second)+","+unparse_annotated(p[1].second)+"]";
+        vector<string> decls;
+        for(auto& [name, value]: p[0].second)
+            decls.push_back(name + " " + show_model_annotated(value));
+        return unparse_annotated(p[1].second) + " where {" + join(decls, "; ") + "}";
     }
     else if (s == "function")
     {
-	return "function["+unparse_annotated(p[0].second)+","+unparse_annotated(p[1].second)+"]";
+	return "|"+unparse_annotated(p[0].second)+":"+unparse_annotated(p[1].second)+"|";
     }
-    if (s == "sample")
+    else if (s == "List")
+    {
+        bool list_of_pairs = true;
+        vector<string> pairs;
+        for(auto& [_,item]: p)
+        {
+            auto type = item.get_child("type");
+            if (type.get_value<string>() == "Tuple" and type.size() == 2)
+            {
+                auto value = item.get_child("value");
+                pairs.push_back(unparse_annotated(value[0].second)+": "+unparse_annotated(value[1].second));
+            }
+            else
+            {
+                list_of_pairs = false;
+                break;
+            }
+        }
+        if (list_of_pairs)
+        {
+            return "{" + join(pairs,", ") + "}";
+        }
+
+        vector<string> items;
+        for(auto& [_,item]: p)
+            items.push_back(unparse_annotated(item));
+        return "[" + join(items,", ") + "]";
+    }
+    else if (s == "Tuple")
+    {
+        vector<string> items;
+        for(auto& [_,item]: p)
+            items.push_back(unparse_annotated(item));
+        return "(" + join(items,", ") + ")";
+    }
+    else if (s == "sample")
 	return "~" + unparse_annotated(p.begin()->second);
-    if (s == "intToDouble")
+    else if (s == "negate")
+	return "-" + unparse_annotated(p.begin()->second);
+    else if (is_operator(s) and p.size() == 2)
+    {
+	// sometimes we might need parenthesis, right?
+        return unparse_annotated(p[0].second) + s + unparse_annotated(p[1].second);
+    }
+
+    else if (s == "intToDouble")
 	return unparse_annotated(p.get_child("x"));
-    if (s == "unit_mixture" or s == "multiMixtureModel")
+    else if (s == "unit_mixture" or s == "multiMixtureModel")
 	if (auto child = p.get_child_optional("submodel"))
 	    return unparse_annotated(*child);
     vector<string> args;
     optional<string> submodel;
+    bool positional = true;
     for(const auto& [arg_name, arg]: p)
     {
 	if (arg.is_null())
-	{
-	    args.push_back("");
-	    continue;
-	}
+            positional = false;
+
 	// Don't print submodel arguments: move out to submodel + <this>
 	else if (arg_name == "submodel")
 	{
+            positional = false;
 	    assert(not submodel);
 	    submodel = unparse_annotated(arg);
-	    args.push_back("");
-            continue;
 	}
 
-        auto arg_value = arg.get_child("value");
-	if (arg_value.has_value<string>() and arg_value.get_value<string>() == "get_state" and arg.get_child("is_default_value") == true)
-	{
-	    args.push_back("");
-	    continue;
-	}
-	else if (arg.get_child_optional("suppress_default") and arg.get_child("suppress_default")== true and arg.get_child("is_default_value") == true)
-	{
-	    args.push_back("");
-	    continue;
-	}
-	else
+	else if (arg.get_child("value").has_value<string>() and arg.get_child("value").get_value<string>() == "get_state" and arg.get_child("is_default_value") == true)
+            positional = false;
+	else if (auto suppress = arg.get_child_optional("suppress_default"); suppress and (*suppress == true) and arg.get_child("is_default_value") == true)
+            positional = false;
+	else if (positional)
 	    args.push_back( unparse_annotated(arg) );
+        else if (auto arg2 = peel_sample_annotated(arg))
+            args.push_back( arg_name + "~" + unparse_annotated(*arg2));
+        else
+            args.push_back( arg_name + "=" + unparse_annotated(arg));
     }
     while (args.size() and args.back() == "")
 	args.pop_back();
     if (not args.empty())
-	s = s + "[" + join(args,',') + "]";
+	s = s + "(" + join(args,", ") + ")";
     if (submodel)
-	s = *submodel + "+" + s;
+	s = *submodel + " +> " + s;
     return s;
 }
 
 string unparse_type(const ptree& p)
 {
-    string s = p.get_value<string>();
-    vector<string> args;
-    for(const auto& pair: p)
-	args.push_back( unparse_type(pair.second) );
-    if (not args.empty())
-	s = s + "[" + join(args,',') + "]";
-    return s;
-}
+    if (p.is_null()) return "NOTYPE";
 
-optional<ptree> peel_sample(ptree p)
-{
-    if (p.has_value<string>() and p.get_value<string>() == "sample")
-	return p[0].second;
-    else
-	return {};
-}
+    auto [head,args] = get_type_apps(p);
 
-optional<ptree> peel_sample_annotated(ptree p)
-{
-    auto v = p.get_child("value");
-    if (v.has_value<string>() and v.get_value<string>() == "sample")
-	return v[0].second;
+    vector<string> sargs;
+    for(const auto& arg: args)
+	sargs.push_back( unparse_type(arg) );
+
+    if (head == "Tuple")
+	return "("+join(sargs,",")+")";
+    else if (head == "Function")
+    {
+	assert(sargs.size() == 2);
+
+	return sargs[0] + " -> " + sargs[1];
+    }
+    else if (not args.empty())
+	return head.get_value<string>() + "<" + join(sargs,',') + ">";
     else
-	return {};
+	return head;
 }
 
 string unparse_abbrev(ptree p, int length)
@@ -628,7 +586,7 @@ string unparse_abbrev(ptree p, int length)
     {
 	output = convertToString(p.value);
 	if (p.size())
-	    output += "[..]";
+	    output += "(..)";
     }
     return output;
 }
@@ -672,3 +630,34 @@ bool is_constant(const ptree& model)
 
     return (name.size()>=2 and name[0] == '"' and name.back() == '"');
 }
+
+int add_arg_placeholder(ptree& p1, const ptree& arg)
+{
+    int n_placeholders = 0;
+
+    for(auto& [key,value]: p1)
+    {
+	if (value == "_")
+	{
+	    n_placeholders++;
+	    value = arg;
+	}
+    }
+
+    return n_placeholders;
+}
+
+
+ptree add_arg(const ptree& p1, const ptree& p2)
+{
+    ptree p3 = p2;
+
+    int n_placeholders = add_arg_placeholder(p3, p1);
+    if (n_placeholders > 1)
+	throw myexception()<<"Placeholder '_' may only occur once.";
+
+    if (n_placeholders == 0)
+	p3.insert(p3.begin(), {"",p1});
+
+    return p3;
+}
