@@ -23,9 +23,8 @@
 #include "util/cmdline.H"
 #include <range/v3/algorithm/max.hpp>
 
-
-using namespace std;
-
+using std::vector;
+using std::string;
 
 sequence_info::sequence_info(const string& n)
     :name(n)
@@ -129,7 +128,7 @@ string guess_alphabet(const vector<sequence>& sequences)
     double digits = letter_fraction("0123456789","-?X=",sequences);
     // FIXME - We can check the largest number ... but each column might have a different highest number.
     if (digits > 0.95)
-	return "Numeric[2]";
+	return "Numeric(2)";
 
     double aa = letter_fraction("ARNDCQEGHILKMFPSTWYVX","-?=",sequences);
     if (letter_fraction("ARNDCQEGHILKMFPSTWYVX","-?",sequences) > 0.9 and AUTGCN<0.5)
@@ -157,26 +156,32 @@ string guess_alphabet(const string& name_, const vector<sequence>& sequences)
 	return guess_alphabet(sequences);
 
     string name = name_;
-    vector<string> arguments = get_arguments(name,'[',']');
+    vector<string> arguments = get_arguments(name,'(',')');
 
     if (name == "Codons")
     {
 	if (arguments.size() < 2) arguments.resize(2);
 	if (arguments[0].empty()) arguments[0] = guess_nucleotides_for(name_, sequences);
 	if (arguments[1].empty()) arguments[1] = "standard";
-	return "Codons[" + arguments[0] + "," + arguments[1] + "]";
+	return "Codons(" + arguments[0] + "," + arguments[1] + ")";
     }
     else if (name == "Doublets")
     {
 	if (arguments.size() < 1) arguments.resize(1);
 	if (arguments[0].empty()) arguments[0] = guess_nucleotides_for(name_, sequences);
-	return "Doublets[" + arguments[0] + "]";
+	return "Doublets(" + arguments[0] + ")";
+    }
+    else if (name == "RNAEdits")
+    {
+	if (arguments.size() < 1) arguments.resize(1);
+	if (arguments[0].empty()) arguments[0] = guess_nucleotides_for(name_, sequences);
+	return "RNAEdits(" + arguments[0] + ")";
     }
     else if (name == "Triplets")
     {
 	if (arguments.size() < 1) arguments.resize(1);
 	if (arguments[0].empty()) arguments[0] = guess_nucleotides_for(name_, sequences);
-	return "Triplets[" + arguments[0] + "]";
+	return "Triplets(" + arguments[0] + ")";
     }
     else
 	return name_;
