@@ -2,7 +2,7 @@ module Test where
 {
   import PopGen;
   import Probability;
-  get_observed_alleles file = map list_from_vector $ list_from_vector $ read_phase_file file;
+  get_observed_alleles file = map vectorToList $ vectorToList $ read_phase_file file;
 
   filename = "/home/bredelings/Reports/Kmar/DC.phase1.infile";
 
@@ -13,7 +13,7 @@ note sigmaOverMu ~ iid(n, gamma(1.05,0.1) );
 
   alpha = 1.0;
 
-note p ~ dirichlet' (n, alpha/(intToDouble n));
+note p ~ symmetric_dirichlet (n, alpha/(intToDouble n));
 
   data1 = get_observed_alleles filename;
 
